@@ -4,6 +4,7 @@ import datetime
 from enum import IntEnum
 import re
 import struct
+import math
 from typing import List, NamedTuple
 
 from bleak import BleakClient
@@ -22,15 +23,27 @@ class Param(IntEnum):
     HUMIDITY = 2
     PRESSURE = 3
     CO2 = 4
-
+    HUMIDITY2 = 5
+    PULSES = 6
+    RADIATION_DOSE = 7
+    RADIATION_DOSE_RATE = 8
+    RADIATION_DOSE_INTEGRAL = 9
 
 class Status(IntEnum):
     """Enum for the different status colors"""
 
+    NONE = 0
     GREEN = 1
     AMBER = 2
     RED = 3
+    BLUE = 4
 
+class AranetType(IntEnum):
+    """Enum for the different Aranet devices"""
+    ARANET4 = 0
+    ARANET2 = 1
+    ARANET_RADIATION = 2
+    UNKNOWN = 255
 
 @dataclass
 class Aranet4HistoryDelegate:
@@ -81,18 +94,111 @@ class CurrentReading:
 
     name: str = ""
     version: str = ""
+    type: AranetType = AranetType.UNKNOWN
+
+    # Aranet2, Aranet4
     temperature: float = -1
-    humidity: int = -1
+    humidity: float = -1
+
+    # Aranet 4
     pressure: float = -1
     co2: int = -1
+
+    # Aranet Radiation
+    radiation_rate: float = -1
+    radiation_total: float = -1
+    radiation_duration: int = -1
+
     battery: int = -1
     status: int = -1
+    status_t: int = -1
+    status_h: int = -1
     interval: int = -1
     ago: int = -1
     stored: int = -1
+    counter: int = -1
+
+    def toString(self, advertisement=None):
+        ret = "=======================================\n"
+        ret += f"  Name:     {self.name}\n"
+
+        if advertisement:
+            ret += f"  Address:  {advertisement.device.address}\n"
+            ret += f"  RSSI:     {advertisement.rssi} dBm\n"
+
+        if self.stored > 0:
+            ret += f"  Logs:     {self.stored}\n"
+
+        ret += "---------------------------------------\n"
+
+        if self.type == AranetType.ARANET4:
+            ret += f"  CO2:            {self.co2} ppm\n"
+            ret += f"  Temperature:    {self.temperature:.01f} \u00b0C\n"
+            ret += f"  Humidity:       {self.humidity} %\n"
+            ret += f"  Pressure:       {self.pressure:.01f} hPa\n"
+            ret += f"  Battery:        {self.battery} %\n"
+            ret += f"  Status Display: {self.status.name}\n"
+            ret += f"  Age:            {self.ago}/{self.interval}\n"
+        elif self.type == AranetType.ARANET2:
+            ret += f"  Temperature:    {self.temperature:.01f} \u00b0C\n"
+            ret += f"  Humidity:       {self.humidity} %\n"
+            ret += f"  Battery:        {self.battery} %\n"
+            ret += f"  Status Temp.:   {self.status_t.name}\n"
+            ret += f"  Status Humid.:  {self.status_h.name}\n"
+            ret += f"  Age:            {self.ago}/{self.interval}\n"
+        elif self.type == AranetType.ARANET_RADIATION:
+            m = math.floor(self.radiation_duration / 60 % 60)
+            h = math.floor(self.radiation_duration / 3600 % 24)
+            d = math.floor(self.radiation_duration / 86400)
+
+            dose_duration = str(m) + "m"
+            if h > 0: dose_duration = str(h) + "h " + dose_duration
+            if d > 0: dose_duration = str(d) + "d " + dose_duration
+
+            ret += f"  Dose rate:      {self.radiation_rate/1000:.02f} uSv/h\n"
+            ret += f"  Dose total:     {self.radiation_total/1000000:.04f} mSv/{dose_duration}\n"
+            ret += f"  Battery:        {self.battery} %\n"
+            ret += f"  Age:            {self.ago}/{self.interval}\n"
+        else:
+            ret += f"  Unknown device type\n"
+
+        return ret
+
+    def toDict(self):
+        data = {
+            "battery": self.battery,
+            "type": self.type.name
+        }
+
+        if self.type == AranetType.ARANET2:
+            data["temperature"] = self.temperature
+            data["humidity"] = self.humidity
+        elif self.type == AranetType.ARANET4:
+            data["co2"] = self.co2
+            data["temperature"] = self.temperature
+            data["pressure"] = self.pressure
+            data["humidity"] = self.humidity
+        elif self.type == AranetType.ARANET_RADIATION:
+            data["radiation_rate"] = self.radiation_rate
+            data["radiation_total"] = self.radiation_total
+            data["radiation_duration"] = self.radiation_duration
+
+        return data
+
+    def decode(self, value: tuple, type: AranetType, gatt=False):
+        """Process advertisement or gatt data"""
+
+        self.type = type
+        if type == AranetType.ARANET4:
+            self._decode_aranet4(value, gatt)
+        elif type == AranetType.ARANET2:
+            self._decode_aranet2(value, gatt)
+        elif type == AranetType.ARANET_RADIATION:
+            self._decode_aranetR(value, gatt)
+
+    def _decode_aranet4(self, value: tuple, gatt=False):
+        """Process Aranet4 data - CO2, Temperature, Humidity, Pressure"""
 
-    def decode(self, value: tuple):
-        """Process data from current log_size and process before storing"""
         self.co2 = self._set(Param.CO2, value[0])
         self.temperature = self._set(Param.TEMPERATURE, value[1])
         self.pressure = self._set(Param.PRESSURE, value[2])
@@ -104,6 +210,62 @@ class CurrentReading:
             self.interval = value[6]
             self.ago = value[7]
 
+    def _decode_aranet2(self, value: tuple, gatt=False):
+        """Process Aranet2 data - Temperature, Humidity"""
+
+        # order from gatt and advertisements are different
+        if gatt:
+            self.temperature = self._set(Param.TEMPERATURE, value[4])
+            self.humidity = self._set(Param.HUMIDITY2, value[5])
+            self.battery = value[3]
+            self.status_h, self.status_t = self._decode_status_flags(value[6])
+            self.interval = value[1]
+            self.ago = value[2]
+        else:
+            self.temperature = self._set(Param.TEMPERATURE, value[1])
+            self.humidity = self._set(Param.HUMIDITY2, value[3])
+            self.status_h, self.status_t = self._decode_status_flags(value[6])
+            self.battery = value[5]
+            self.interval = value[7]
+            self.ago = value[8]
+            self.counter = value[9]
+
+    def _decode_aranetR(self, value: tuple, gatt=False):
+        """Process Aranet Radiation data - radiation dose"""
+        # order from gatt and advertisements are different
+        if gatt:
+            self.radiation_duration = value[6]
+            self.radiation_rate = value[4]
+            self.radiation_total = value[5]
+            self.battery = value[3]
+            self.interval = value[1]
+            self.ago = value[2]
+        else:
+            self.radiation_total = value[0]
+            self.radiation_duration = value[1]
+            self.radiation_rate = value[2]
+            self.battery = value[4]
+            self.interval = value[6]
+            self.ago = value[7]
+            self.counter = value[8]
+
+    @staticmethod
+    def _decode_status_flags(status):
+        status_t = Status.GREEN
+        status_h = Status.GREEN
+
+        if status & 0b0001:
+            status_h = Status.BLUE
+        elif status & 0b0010:
+            status_h = Status.RED
+
+        if status & 0b0100:
+            status_t = Status.BLUE
+        elif status & 0b1000:
+            status_t = Status.RED
+
+        return (status_h, status_t)
+
     @staticmethod
     def _set(param: Param, value: int):
         """
@@ -125,6 +287,18 @@ class CurrentReading:
         elif param == Param.HUMIDITY:
             invalid_reading_flag = value >> 8
             multiplier = 1
+        elif param == Param.HUMIDITY2:
+            invalid_reading_flag = value >> 15 == 1
+            multiplier = 0.1
+        elif param == Param.RADIATION_DOSE:
+            invalid_reading_flag = value >> 15 == 1
+            multiplier = 1 # nSv
+        elif param == Param.RADIATION_DOSE_RATE:
+            invalid_reading_flag = value >> 15 == 1
+            multiplier = 10 # nSv/h
+        elif param == Param.RADIATION_DOSE_INTEGRAL:
+            invalid_reading_flag = value >> 15 == 1
+            multiplier = 1 # nSv
 
         if invalid_reading_flag:
             return -1
@@ -133,7 +307,7 @@ class CurrentReading:
         return value * multiplier
 
 
-@dataclass
+@dataclass(order=True)
 class Version:
     major: int = -1
     minor: int = -1
@@ -187,6 +361,52 @@ class Aranet4Advertisement:
     device: BLEDevice = None
     readings: CurrentReading = None
     manufacturer_data: ManufacturerData = None
+    rssi: int = None
+
+    def __init__(self, device = None, ad_data = None):
+        self.device = device
+
+        if device and ad_data:
+            has_manufacurer_data = Aranet4.MANUFACTURER_ID in ad_data.manufacturer_data
+            self.rssi = ad_data.rssi
+
+            if has_manufacurer_data:
+                mf_data = ManufacturerData()
+                raw_bytes = bytearray(ad_data.manufacturer_data[Aranet4.MANUFACTURER_ID])
+
+                if (not device.name and len(raw_bytes) in [7,22]) or "Aranet4" in device.name:
+                    raw_bytes.insert(0,0)
+
+                # Basic info
+                value_fmt = "<BBBB"
+                value = struct.unpack(value_fmt, raw_bytes[1:5])
+                mf_data.decode(value)
+                self.manufacturer_data = mf_data
+
+                # Extended info / measurements
+                aranetv = raw_bytes[0]
+
+                if aranetv == 0: # Aranet4
+                    value_fmt = "<xxxxxxxxxHHHBBBHH"
+                    aranetv = AranetType.ARANET4
+                elif aranetv == 1: # Aranet2
+                    value_fmt = "<xxxxxxxxHHHHBBBHHB"
+                    aranetv = AranetType.ARANET2
+                elif aranetv == 2: # Aranet Radiation
+                    value_fmt = "<xxxxxxIIHBBBHHB"
+                    aranetv = AranetType.ARANET_RADIATION
+                else:
+                    value_fmt = ""
+                    aranetv = None
+
+                end = struct.calcsize(value_fmt)
+                if end > 0 and len(raw_bytes[:end]) == end:
+                    value = struct.unpack(value_fmt, raw_bytes[:end])
+                    self.readings = CurrentReading()
+                    self.readings.decode(value, aranetv)
+                    self.readings.name = device.name
+                else:
+                    mf_data.integrations = False
 
 
 @dataclass
@@ -198,6 +418,9 @@ class RecordItem:
     humidity: int
     pressure: float
     co2: int
+    rad_dose: float
+    rad_dose_rate: float
+    rad_dose_total: float
 
 
 @dataclass
@@ -207,9 +430,12 @@ class Filter:
     begin: int
     end: int
     incl_temperature: bool
-    incl_humidity: bool
+    incl_humidity: int
     incl_pressure: bool
     incl_co2: bool
+    incl_rad_dose: bool
+    incl_rad_dose_rate: bool
+    incl_rad_dose_total: bool
 
 
 @dataclass
@@ -252,6 +478,9 @@ class Aranet4:
     # Read / Aranet service
     AR4_READ_CURRENT_READINGS = "f0cd1503-95da-4f4b-9ac8-aa55d312af0c"
     AR4_READ_CURRENT_READINGS_DET = "f0cd3001-95da-4f4b-9ac8-aa55d312af0c"
+    AR2_READ_CURRENT_READINGS = "f0cd1504-95da-4f4b-9ac8-aa55d312af0c"
+    AR2_READ_CURRENT_READINGS_A = "f0cd3003-95da-4f4b-9ac8-aa55d312af0c" # data is same as other
+    # aranet2 has different service uids. use rhose
     AR4_READ_INTERVAL = "f0cd2002-95da-4f4b-9ac8-aa55d312af0c"
     AR4_READ_SECONDS_SINCE_UPDATE = "f0cd2004-95da-4f4b-9ac8-aa55d312af0c"
     AR4_READ_TOTAL_READINGS = "f0cd2001-95da-4f4b-9ac8-aa55d312af0c"
@@ -286,6 +515,11 @@ class Aranet4:
         self.device = BleakClient(address)
         self.reading = True
 
+    def __del__(self):
+        """Close remote"""
+        if self.device.is_connected:
+            asyncio.shield(self.device.disconnect())
+
     async def connect(self):
         """Connect to remote device"""
         await self.device.connect()
@@ -294,18 +528,37 @@ class Aranet4:
         """Extract current readings from remote device"""
         readings = CurrentReading()
 
-        if details:
-            uuid = self.AR4_READ_CURRENT_READINGS_DET
-            # co2, temp, pressure, humidity, battery, status
-            value_fmt = "<HHHBBBHH"
-        else:
-            uuid = self.AR4_READ_CURRENT_READINGS
-            # co2, temp, pressure, humidity, battery, status, interval, ago
-            value_fmt = "<HHHBBB"
+        aranet2_char = self.device.services.get_characteristic(
+            self.AR2_READ_CURRENT_READINGS
+        )
 
-        raw_bytes = await self.device.read_gatt_char(uuid)
-        value = struct.unpack(value_fmt, raw_bytes)
-        readings.decode(value)
+        if aranet2_char:
+            uuid = self.AR2_READ_CURRENT_READINGS
+            raw_bytes = await self.device.read_gatt_char(uuid)
+
+            # TODO: Can we use byte ar position 0 to detect type?
+            if len(raw_bytes) >= 28:
+                # radiation
+                value_fmt = "<HHHBIQQB"
+                value = struct.unpack(value_fmt, raw_bytes[0:28])
+                readings.decode(value, AranetType.ARANET_RADIATION, True)
+            else:
+                value_fmt = "<HHHBHHB"
+                value = struct.unpack(value_fmt, raw_bytes)
+                readings.decode(value, AranetType.ARANET2, True)
+        else:
+            if details:
+                uuid = self.AR4_READ_CURRENT_READINGS_DET
+                # co2, temp, pressure, humidity, battery, status
+                value_fmt = "<HHHBBBHH"
+            else:
+                uuid = self.AR4_READ_CURRENT_READINGS
+                # co2, temp, pressure, humidity, battery, status, interval, ^
+                value_fmt = "<HHHBBB"
+
+            raw_bytes = await self.device.read_gatt_char(uuid)
+            value = struct.unpack(value_fmt, raw_bytes)
+            readings.decode(value, AranetType.ARANET4)
         return readings
 
     async def get_interval(self) -> int:
@@ -398,14 +651,29 @@ class Aranet4:
             )
 
             header = HistoryHeader(*struct.unpack("<BHHHHB", packet[:10]))
+            packet = packet[10:]
 
             if header.param != param.value or header.count == 0:
                 await asyncio.sleep(0.1)
                 continue
 
-            pattern = "<B" if param.value == Param.HUMIDITY else "<H"
+            if param.value == Param.HUMIDITY:
+                pattern = "<B"
+            elif param.value == Param.RADIATION_DOSE:
+                pattern = "<Hx"
+            elif param.value == Param.RADIATION_DOSE_RATE:
+                pattern = "<Hx"
+            elif param.value == Param.RADIATION_DOSE_INTEGRAL:
+                pattern = "<Q"
+            else:
+                pattern = "<H"
 
-            data_values = struct.iter_unpack(pattern, packet[10:])
+            # trim to multiples of pattern size
+            plen = len(packet)
+            packet = packet[:plen - plen % struct.calcsize(pattern)]
+
+            # extract values
+            data_values = struct.iter_unpack(pattern, packet)
             for idx, value in enumerate(data_values, header.start - 1):
                 if idx > end or idx == header.start - 1 + header.count:
                     break
@@ -450,6 +718,27 @@ class Aranet4:
         await self.device.stop_notify(self.AR4_READ_HISTORY_READINGS_V1)
         return delegate.result
 
+    async def set_readings_interval(self, interval: int):
+        """Set reading interval"""
+        header = 0x90
+        val = struct.pack("<BB", header, interval)
+        await self.device.write_gatt_char(self.AR4_WRITE_CMD, val, True)
+
+    async def set_home_integration_enabled(self, enabled: bool):
+        """
+        Toggle smart home integrations.
+        This is required to receive measurements in advertisements.
+        """
+        header = 0x91
+        val = struct.pack("<BB", header, 1 if enabled else 0)
+        await self.device.write_gatt_char(self.AR4_WRITE_CMD, val, True)
+
+    async def set_bluetooth_range(self, extended: bool):
+        """Set bluetooth range"""
+        header = 0x91
+        val = struct.pack("<BB", header, 1 if extended else 0)
+        await self.device.write_gatt_char(self.AR4_WRITE_CMD, val, True)
+
 
 def _log_times(now, total, interval, ago):
     """Calculate the actual times datapoints were logged on device"""
@@ -528,32 +817,7 @@ class Aranet4Scanner:
 
     def _process_advertisement(self, device, ad_data):
         """Processes Aranet4 advertisement data"""
-
-        adv = Aranet4Advertisement()
-        adv.device = device
-
-        has_manufacurer_data = Aranet4.MANUFACTURER_ID in ad_data.manufacturer_data
-        has_service = Aranet4.AR4_SERVICE in ad_data.service_data
-
-        if has_manufacurer_data:
-            mf_data = ManufacturerData()
-            raw_bytes = ad_data.manufacturer_data[Aranet4.MANUFACTURER_ID]
-
-            # Basic info
-            value_fmt = "<BBBHBB"
-            value = struct.unpack(value_fmt, raw_bytes[0:7])
-            mf_data.decode(value)
-            adv.manufacturer_data = mf_data
-
-            # Extended info / measurements
-            if len(raw_bytes) >= 20:
-                value_fmt = "<HHHBBBHH"
-                value = struct.unpack(value_fmt, raw_bytes[8:21])
-                adv.readings = CurrentReading()
-                adv.readings.decode(value)
-                adv.readings.name = device.name
-
-
+        adv = Aranet4Advertisement(device, ad_data)
         self.on_scan(adv)
 
     def __init__(self, on_scan):
@@ -577,7 +841,7 @@ async def _find_nearby(detect_callback: callable, duration: int) -> List[BLEDevi
     await scanner.stop()
     return [device
             for device in scanner.scanner.discovered_devices
-            if "Aranet4" in device.name]
+            if "Aranet" in device.name]
 
 
 def find_nearby(detect_callback: callable, duration: int = 5) -> List[BLEDevice]:
@@ -620,6 +884,20 @@ async def _all_records(address, entry_filter, remove_empty):
         last_log = await monitor.get_seconds_since_update()
         now = datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0)
 
+    if dev_name.startswith("Aranet2"):
+        entry_filter["pres"] = False
+        entry_filter["co2"] = False
+        if entry_filter.get("humi", False):
+            entry_filter["humi"] = 2 #v2 humidity
+    elif dev_name.startswith("Aranet\u2622"):
+        entry_filter["pres"] = False
+        entry_filter["co2"] = False
+        entry_filter["humi"] = False
+        entry_filter["temp"] = False
+        entry_filter["rad_dose"] = entry_filter.get("rad_dose", True)
+        entry_filter["rad_dose_rate"] = entry_filter.get("rad_dose_rate", True)
+        entry_filter["rad_dose_total"] = entry_filter.get("rad_dose_total", True)
+
     log_size = await monitor.get_total_readings()
     log_points = _log_times(now, log_size, interval, last_log)
     begin, end = _calc_start_end(log_points, entry_filter)
@@ -630,6 +908,9 @@ async def _all_records(address, entry_filter, remove_empty):
         entry_filter.get("humi", True),
         entry_filter.get("pres", True),
         entry_filter.get("co2", True),
+        entry_filter.get("rad_dose", False),
+        entry_filter.get("rad_dose_rate", False),
+        entry_filter.get("rad_dose_total", False),
     )
 
     if begin < 0 or end < 0:
@@ -643,7 +924,11 @@ async def _all_records(address, entry_filter, remove_empty):
         )
     else:
         temperature_val = _empty_reading(log_size)
-    if rec_filter.incl_humidity:
+    if rec_filter.incl_humidity == 2:
+        humidity_val = await monitor.get_records(
+            Param.HUMIDITY2, log_size=log_size, start=begin, end=end
+        )
+    elif rec_filter.incl_humidity:
         humidity_val = await monitor.get_records(
             Param.HUMIDITY, log_size=log_size, start=begin, end=end
         )
@@ -661,11 +946,30 @@ async def _all_records(address, entry_filter, remove_empty):
         )
     else:
         co2_val = _empty_reading(log_size)
+    if rec_filter.incl_rad_dose:
+        rad_dose_val = await monitor.get_records(
+            Param.RADIATION_DOSE, log_size=log_size, start=begin, end=end
+        )
+    else:
+        rad_dose_val = _empty_reading(log_size)
+    if rec_filter.incl_rad_dose_rate:
+        rad_dose_rate_val = await monitor.get_records(
+            Param.RADIATION_DOSE_RATE, log_size=log_size, start=begin, end=end
+        )
+    else:
+        rad_dose_total_val = _empty_reading(log_size)
+    if rec_filter.incl_rad_dose_total:
+        rad_dose_total_val = await monitor.get_records(
+            Param.RADIATION_DOSE_INTEGRAL, log_size=log_size, start=begin, end=end
+        )
+    else:
+        rad_dose_total_val = _empty_reading(log_size)
+####
     # Store returned data in dataclass
-    data = zip(log_points, co2_val, temperature_val, pressure_val, humidity_val)
+    data = zip(log_points, co2_val, temperature_val, pressure_val, humidity_val, rad_dose_val, rad_dose_rate_val, rad_dose_total_val)
     record = Record(dev_name, dev_version, log_size, rec_filter)
-    for date, co2, temp, pres, hum in data:
-        record.value.append(RecordItem(date, temp, hum, pres, co2))
+    for date, co2, temp, pres, hum, rad, rad_rate, rad_integral in data:
+        record.value.append(RecordItem(date, temp, hum, pres, co2, rad, rad_rate, rad_integral))
     if (remove_empty):
         record.value = record.value[begin-1:end+1]
     return record
